package org.openjdk.varhandles;

import org.openjdk.jmh.annotations.*;
import org.openjdk.jmh.infra.Blackhole;
import sun.misc.Unsafe;

import java.lang.invoke.MethodHandles;
import java.lang.invoke.VarHandle;
import java.lang.reflect.Field;
import java.util.concurrent.TimeUnit;

@Warmup(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Measurement(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Fork(3)
@State(Scope.Benchmark)
@BenchmarkMode(Mode.AverageTime)
@OutputTimeUnit(TimeUnit.NANOSECONDS)
public class CAEvsCAS {

    public static final Unsafe U;
    public static final long OFF;
    public static final VarHandle VH;

    static {
        try {
            Field unsafeField = Unsafe.class.getDeclaredField("theUnsafe");
            unsafeField.setAccessible(true);
            U = (Unsafe) unsafeField.get(null);
            OFF = U.objectFieldOffset(CAEvsCAS.class.getDeclaredField("v"));
        } catch (Exception e) {
            throw new AssertionError(e);
        }

        try {
            MethodHandles.Lookup lookup = MethodHandles.lookup().in(CAEvsCAS.class);
            VH = lookup.findVarHandle(CAEvsCAS.class, "v", int.class);
        } catch (Exception e) {
            throw new IllegalStateException(e);
        }

    }

    private int v;

    @Param({"0", "1", "4", "16", "64", "256", "1024", "2048", "4096"})
    private int backoff;

    private int mutate(int v) {
        return (v + 1) & 0xFFFF;
    }

    @Benchmark
    @CompilerControl(CompilerControl.Mode.DONT_INLINE)
    public int swap() {
        Blackhole.consumeCPU(backoff);
        int ex, nx;
        do {
            ex = U.getIntVolatile(this, OFF);
            nx = mutate(ex);
        } while(!VH.compareAndSet(this, ex, nx));
        return nx;
    }

    @Benchmark
    @CompilerControl(CompilerControl.Mode.DONT_INLINE)
    public int exchange() {
        Blackhole.consumeCPU(backoff);
        int x = U.getIntVolatile(this, OFF);
        int ex, nx;
        do {
            ex = x;
            nx = mutate(ex);
            x = (int) VH.compareAndExchangeVolatile(this, ex, nx);
        } while (x != ex);
        return nx;
    }

    @Benchmark
    @CompilerControl(CompilerControl.Mode.DONT_INLINE)
    public int exchange_withReread() {
        Blackhole.consumeCPU(backoff);
        int x;
        int ex, nx;
        do {
            ex = U.getIntVolatile(this, OFF);
            nx = mutate(ex);
            x = (int) VH.compareAndExchangeVolatile(this, ex, nx);
        } while (x != ex);
        return nx;
    }

}
